#include <iostream>
#include <fstream>

#include "Vec3D.h"
#include "BoundingBox.h"
#include "Polygonizer/polygonizer.h"


using namespace std;



// Mesh Structure

class Vertex {
public:
    inline Vertex () {}
    inline Vertex (const Vec3Df & p) : p (p) {}
    inline Vertex (const Vertex & v) : p (v.p) {}
    inline virtual ~Vertex () {}
    inline Vertex & operator= (const Vertex & v) {
        p = v.p;
        return (*this);
    }
    Vec3Df p;
};

class Triangle {
public:
    inline Triangle () {
        v[0] = v[1] = v[2] = 0;
    }
    inline Triangle (const Triangle & t) {
        v[0] = t.v[0];
        v[1] = t.v[1];
        v[2] = t.v[2];
    }
    inline Triangle (unsigned int v0, unsigned int v1, unsigned int v2) {
        v[0] = v0;
        v[1] = v1;
        v[2] = v2;
    }
    inline virtual ~Triangle () {}
    inline Triangle & operator= (const Triangle & t) {
        v[0] = t.v[0];
        v[1] = t.v[1];
        v[2] = t.v[2];
        return (*this);
    }
    unsigned int v[3];
};

class Mesh {
public:
    vector<Vertex> V;
    vector<Triangle> T;

    void saveOFF (const string & filename);
};


void Mesh::saveOFF (const string & filename) {
    ofstream out (filename.c_str ());
    if (!out) 
        exit (EXIT_FAILURE);
    out << "OFF" << endl;
    out << V.size () << " " << T.size () << " 0" << endl;
    for (unsigned int i = 0; i < V.size (); i++)
        out << V[i].p[0] << " " << V[i].p[1] << " " << V[i].p[2] << endl;
    for (unsigned int i = 0; i < T.size (); i++)
        out << "3 " << T[i].v[0] << " " << T[i].v[1] << " " << T[i].v[2] << endl;
    out.close ();
}

// Grid and Implicit Function

class Grid: public ImplicitFunction {
public:
    unsigned int resolution;
    float size;
    vector<bool> values;

    Polygonizer * pol;

    inline Grid (unsigned int resolution) : resolution (resolution) {
        values.resize (resolution*resolution*resolution, false);
    }

    inline virtual ~Grid() {
        delete pol;
    }
    
    
    void open (const string & filename) {
        ifstream in (filename.c_str ());
        if (!in) 
            exit (EXIT_FAILURE);
        unsigned int resX, resY, resZ;
        float cellX, cellY, cellZ;
        in >> cellX >> cellY >> cellZ >> resX >> resY >> resZ;
       
        float sizeX = cellX * resX;
        float sizeY = cellY * resY;
        float sizeZ = cellZ * resZ;
        size = max (sizeX, max (sizeY, sizeZ));
        
        unsigned int count = resX * resY * resZ;
        unsigned int x = 0;
        unsigned int y = 0;
        unsigned int z = 0;
        for (unsigned int i = 0; i < count; i++) {
            unsigned int v;
            in >> v;
            if (v != 0) {

                int dx = (resolution*sizeX*x)/(resX*size);
                if (dx < 0) dx = 0;
                if (dx > (resolution - 1)) dx = resolution - 1;
                
                int dy = (resolution*sizeY*y)/(resY*size);
                if (dy < 0) dy = 0;
                if (dy > (resolution - 1)) dy = resolution - 1;
                
                int dz = (resolution*sizeZ*z)/(resZ*size);
                if (dz < 0) dz = 0;
                if (dz > (resolution - 1)) dz = resolution - 1;
                unsigned int index = dx + resolution * (dy + dz * resolution);
                values[index] = true;
            }
            x++;
            if (x == resX) {
                y++;
                if (y == resY) {
                    z++;
                    y = 0;
                }
                x = 0;
            }
        }
    }

    float eval (float x, float y, float z) {
        int dx = x * resolution;
        if (dx < 0 || dx >= resolution)
            return -1.0;
        int dy = y * resolution;
        if (dy < 0 || dy >= resolution)
            return -1.0;
        int dz = z * resolution;
        if (dz < 0 || dz >= resolution)
            return -1.0;
        unsigned int it = dx + resolution * (dy + dz * resolution);
        bool v;
        if (it < values.size ())
            v = values[it];
        else
            v = false;
        if (v)
            return 1.0;
        return -1.0;
    }

    void contour ();
    void extractMesh (Mesh & m);
};

void Grid::contour () {
    Vec3Df initMarchPos;
    for (unsigned int i = 0; i < resolution; i++)
        for (unsigned int j = 0; j < resolution; j++)
            for (unsigned int k = 0; k < resolution; k++) {
                unsigned int it = i + resolution * (j + resolution * k);
                if (values[it]) 
                    initMarchPos = Vec3Df (i, j, k)/resolution;
            }
    pol = new Polygonizer (this, 1.0f/resolution, resolution);
    pol->march (false, initMarchPos[0], initMarchPos[1], initMarchPos[2]);
}

void Grid::extractMesh (Mesh & m) {
    for (int i = 0; i < pol->no_vertices (); i++) {
        VERTEX v = pol->get_vertex (i);
        Vertex vertex (Vec3Df (v.x, v.y, v.z));
        m.V.push_back (vertex);
    }
    for (int i = 0; i < pol->no_triangles(); i++) {
        TRIANGLE t = pol->get_triangle(i);
        m.T.push_back (Triangle (t.v0, t.v1, t.v2));
    }
}

void printUsage () {
    cerr << endl << "grid2mesh - Volumetric grid meshing tool, by Tamy Boubekeur." << endl
         << "-----------------" << endl
         << "Usage: ./grid2mesh <output-res> <input-file.grid> <output-file.off>" << endl
         << "Grid format (ascii):" << endl
         << "<cell X size> <cell Y size> <cell Z size>" << endl
         << "<grid X res.> <grid Y res.> <grid Z res.>" << endl
         << "values/labels ..." << endl << endl;
    
}

int main (int argc, char ** argv) {
    string infilename;
    string outfilename;
    unsigned int outputRes = 128;
    if (argc == 4) {
        outputRes = atoi (argv[1]);
        infilename = string (argv[2]);
        outfilename = string (argv[3]);
    } else  {
        printUsage ();
        exit (EXIT_FAILURE);
    }
    Grid grid (outputRes);
    grid.open (infilename);
    grid.contour ();
    Mesh m;
    grid.extractMesh (m);
    m.saveOFF (outfilename);
    
    return 0;
}
